"""Base Error Class for FedRAG."""


class FedRAGError(Exception):
    """Base error for all fed-rag exceptions."""

    pass


class FedRAGWarning(Warning):
    """Base warning for all fed-rag warnings."""
